from MichaelisMenten import MichaelisMenten


class MichaelisMentenGeneral(MichaelisMenten):

    def __init__(self, forwardSpecie: str, backwardSpecie: str, Vmax_f: float = 10, Km_b: float = 100, Vmax_b: float = 10, Km_f: float = 100, noForward=False, noBackward=False):

        # Generalised forward Michaelis Menten kinetics
        # incorporates stimulation and inhibition
        # need to include kinetic constant values for each stimulator/inhibitor
        # Assumption: each stimulator/inhibitor is independent!
        # NOTE: currently, this class assumes allosteric inhibition 

        super().__init__(forwardSpecie, backwardSpecie, Vmax_f=Vmax_f, Km_b=Km_b, Vmax_b=Vmax_b, Km_f=Km_f)
        self.type = "MG"
        
        self.stimulatorsForward = []
        self.inhibitorsForward = []

        self.stimulatorsBackward = []
        self.inhibitorsBackward = []

        self.noForward = noForward
        self.noBackward = noBackward

    def addStimulator(self, specie: str, kc: float = 0.1, backward=False):

        if backward:
            self.stimulatorsBackward.append((specie, kc))
        else:    
            self.stimulatorsForward.append((specie, kc))

    def addInhibitor(self, specie: str, ki: float = 0.01, backward=False):

        if backward:
            self.inhibitorsBackward.append((specie, ki))
        else:
            self.inhibitorsForward.append((specie, ki))

    def computeForward(self, forwardValues: list):

        # NOTE: not utilised

        return super().computeForward(forwardValues)

        
    def computeBackward(self, backwardValues: list):
        
        # NOTE: not utilised 

        return super().computeBackward(backwardValues)

    def getEqHeaderStr(self, index):
        if self.noBackward:
            return "{forward} => {backward} :R{i}".format(forward=self.fs[0], backward=self.bs[0], i=index)
        return super().getEqHeaderStr(index)

    def getForwardEqStr(self, index):

        if self.noForward:
            return None

        if len(self.stimulatorsForward) == 0 and len(self.inhibitorsForward) == 0:
            return super().getForwardEqStr(index)

        retStr = ""
        top = ""
        bot = ""

        top += "("
        if len(self.stimulatorsForward) == 0:
            top += "Vmax{id}f".format(id=index)
        else:
            i = 0 
            while i < len(self.stimulatorsForward):
                st = self.stimulatorsForward[i][0]
                top = top + "Kc{id}f{kc_id} * {st} + ".format(id=index, kc_id=str(i+1), st=st)
                i += 1  
            top = top[:-3]
        top += ")"
        top += " * {fs}".format(fs=[0])

        bot = "(Km{id}f + {fs}) * ".format(id=index, fs=self.fs[0])
        i = 0
        while i < len(self.inhibitorsForward):
            In = self.inhibitorsForward[i][0]
            bot = bot + "(1 + {In} / Ki{id}f{ki_id}) * ".format(In=In, id=index, ki_id=str(i+1))
            i += 1
        bot = bot[:-3]             

        retStr = top + " / " + bot
        return retStr

    def getBackwardEqStr(self, index):

        if self.noBackward:
            return None

        if len(self.stimulatorsBackward) == 0 and len(self.inhibitorsBackward) == 0:
            return super().getBackwardEqStr(index)

        retStr = ""
        top = ""
        bot = ""

        top += "("
        if len(self.stimulatorsBackward) == 0:
            top += "Vmax{id}b".format(id=index)
        else:
            i = 0 
            while i < len(self.stimulatorsBackward):
                st = self.stimulatorsBackward[i][0]
                top = top + "Kc{id}b{kc_id} * {st} + ".format(id=index, kc_id=str(i+1), st=st)
                i += 1  
            top = top[:-3]
        top += ")"
        top += " * {bs}".format(bs=self.bs[0])

        bot = "(Km{id}f + {bs}) * ".format(id=index, bs=self.bs[0])
        i = 0
        while i < len(self.inhibitorsBackward):
            In = self.inhibitorsBackward[i][0]
            bot = bot + "(1 + {In} / Ki{id}f{ki_id}) * ".format(In=In, id=index, ki_id=str(i+1))
            i += 1
        bot = bot[:-3]             

        retStr = top + " / " + bot


        return retStr


    def getParams(self, index):


        all_params = {}

        if len(self.stimulatorsForward) == 0:
            # return Vmaxf
            vmax_str = "Vmax{id}f".format(id=index)
            all_params[vmax_str] = self.Vmax_f

        if len(self.stimulatorsBackward) == 0:
            # return Vmaxb 
            vmax_str = "Vmax{id}b".format(id=index)
            all_params[vmax_str] = self.Vmax_b
        
        kmf_str = "Km{id}f".format(id=index)
        Kmb_str = "Km{id}b".format(id=index)
        all_params[kmf_str] = self.Km_f
        all_params[Kmb_str] = self.Km_b
        # NOTE: REALLY SHOULD REFACTOR

        i = 0 
        while i < len(self.stimulatorsForward):
            k_str = "Kc{id}f{kc_id}".format(id=index, kc_id=str(i+1))
            all_params[k_str] = self.stimulatorsForward[i][1]
            i += 1

        i = 0
        while i < len(self.stimulatorsBackward):
            k_str = "Kc{id}b{kc_id}".format(id=index, kc_id=str(i+1))
            all_params[k_str] = self.stimulatorsBackward[i][1]
            i += 1

        i = 0
        while i < len(self.inhibitorsForward):
            k_str = "Ki{id}f{ki_id}".format(id=index, ki_id=str(i+1))
            all_params[k_str] = self.inhibitorsForward[i][1]
            i += 1

        i = 0
        while i < len(self.inhibitorsBackward):
            k_str = "Ki{id}f{ki_id}".format(id=index, ki_id=str(i+1))
            all_params[k_str] = self.inhibitorsBackward[i][1]
            i += 1
    
        return all_params


if __name__ == "__main__":
    
    r = MichaelisMentenGeneral("A", "B")
    r.addStimulator("C")
    r.addStimulator("D")
    r.addInhibitor("I", backward=True)
    r.addInhibitor("I2", backward=True)
    print(r.getForwardEqStr(1))
    print(r.getBackwardEqStr(1))
    print(r.getParams(1))



from Reactions import Reactions

class Degradation(Reactions):

    def __init__(self, forwardSpecies, KDeg: float = 0.01):
        super().__init__(forwardSpecies, "None")
        self.KDeg = KDeg
        self.stimulators = []

    def addStimulator(self, specie: str, alpha: float = 0.1):
        
        self.stimulators.append((specie, alpha))

    def getEqHeaderStr(self, index):
        return "{forward} => :R{i}".format(forward=self.fs[0], i=index)

    def getForwardEqStr(self, index):
        retStr = "Kdeg{i} * {fs}".format(i=index, fs=self.fs[0])

        if len(self.stimulators) == 0:
            return "Kdeg{i} * {fs}".format(i=index, fs=self.fs[0])

        addi = " * (1 + "
        i = 0 
        while i < len(self.stimulators):
            st = self.stimulators[i]
            specie = st[0]
            addiEq = "(alpha{i}b{ai} * {st})".format(i=index, ai=str(i+1), st=specie)
            addi = addi + addiEq
            i += 1
            if i != len(self.stimulators):
                addi += " + "
        addi += ")"

        return retStr + addi

    def getBackwardEqStr(self, index):
        return super().getBackwardEqStr(index)

    def getParams(self, index):

        all_params = {"Kdeg{i}".format(i=index): self.KDeg}

        i = 0 
        while i < len(self.stimulators):
            k_str = "alpha{id}b{ai}".format(id=index, ai=str(i+1))
            all_params[k_str] = self.stimulators[i][1]
            i += 1

        return all_params


if __name__ == "__main__":

    r = Degradation("A")
    r.addStimulator("C")
    r.addStimulator("D")

    print(r.getBackwardEqStr(1))
    print(r.getParams(1))



from .Degradation import Degradation
from .MassAction import MassAction
from .MichaelisMenten import MichaelisMenten
from .MichaelisMentenGeneral import MichaelisMentenGeneral
from .Synthesis import Synthesis
from .Reactions import Reactions



class Reactions:

    def __init__(self, forwardSpecies, backwardSpecies):

        # forwardSpecies and backwardSpecies can be list or str
        # but are always initialised as list
        if isinstance(forwardSpecies, str):
            self.fs = [forwardSpecies]
        else:
            self.fs = forwardSpecies
        if isinstance(backwardSpecies, str):
            self.bs = [backwardSpecies]
        else: 
            self.bs = backwardSpecies
        self.type = "None"

    def computeForward(self, forwardValues: list):
        
        return None

    def computeBackward(self, backwardValues: list):

        return None

    def getEqHeaderStr(self, index):

        return None

    def getForwardEqStr(self, index):

        return None

    def getBackwardEqStr(self, index):

        return None

    def getParams(self, index):

        return None


from .Reactions import Reactions

class Synthesis(Reactions):

    def __init__(self, backwardSpecies, KSyn: float = 0.01):
        super().__init__("None", backwardSpecies)
        self.KSyn = KSyn

    def getEqHeaderStr(self, index):
        return " => {backward} :R{i}".format(backward=self.bs[0], i=index)

    def getForwardEqStr(self, index):
        return "Ksyn{i}".format(i=index)

    def getBackwardEqStr(self, index):
        return super().getBackwardEqStr(index)

    def getParams(self, index):
        return {"Ksyn{i}".format(i=index): self.KSyn}




# a new model definition where it is essentially a wrapper
# for a set of reactions

from .reactions.Reactions import Reactions

class Model: 

    def __init__(self, modelName = "None"):

        self.reactions = []
        self.modelName = modelName

        # structure: "specieName": "initial concentration"
        self.activators = []
        self.specieFamily = {}
        self.species = {}

    def setModelName(self, name):

        self.modelName = name

    def updateSpecies(self):

        # checks every reaction within the model
        # a O(N) update

        i = 0 
        while i < len(self.reactions):
            re = self.reactions[i]
            for fs in re.fs:
                if fs not in self.species and fs != 'None' and fs != None:
                    self.species[fs] = 100
            
            for bs in re.bs:
                if bs not in self.species and bs != 'None' and bs != None:
                    self.species[bs] = 100
            i += 1

    def specieConc(self, specieNames: list, conc: list):

        assert len(specieNames) == len(conc)

        i = 0 
        while i < len(specieNames):
            s = specieNames[i]
            self.species[s] = conc[i]
            i += 1

    def addReaction(self, reaction: Reactions):

        self.reactions.append(reaction)
        self.updateSpecies()

    def addActivation(self, activator: str, conc: float, activationTime: float):

        self.activators.append((activator, conc, activationTime))

    def generateTxtbc(self):
        
        txtbc = open("{filename}.txtbc".format(filename=self.modelName), "w")

        ### HEADER

        txtbc.write("********** MODEL NAME\n")
        txtbc.write(self.modelName + "\n")
        txtbc.write("\n")
        txtbc.write("********** MODEL NOTES\n")
        txtbc.write("\n")

        ### Model state information

        txtbc.write("********** MODEL STATE INFORMATION\n")
        txtbc.write("% Initial Conditions\n")
        txtbc.write("\n")

        species = self.species.keys()
        for s in species:
            txtbc.write("{specie}(0) = {conc}\n".format(specie = s, conc = self.species[s]))
        txtbc.write("\n")

        
        # NOTE: deprecated code below, have not deleted for possible later reference

        # families = self.specieFamily.keys()
        # families = list(families)

        # for fam in families:
        #     for member in self.specieFamily[fam]: 
        #         txtbc.write("{specie}(0) = {conc}\n".format(specie = member, conc = self.species[member]))
        #     txtbc.write("\n")


        ### Model parameters

        txtbc.write("********** MODEL PARAMETERS\n")

        i = 0 
        while i < len(self.reactions):
            re = self.reactions[i]
            id_ = i + 1
            re_params = re.getParams(id_)
            names = re_params.keys()
            for n in names:
                toStr = n + " = " + str(re_params[n]) + "\n"
                txtbc.write(toStr)
            
            txtbc.write("\n")
            i += 1

        txtbc.write("% Constants\n")
        txtbc.write("\n")

        txtbc.write("% Stimulation Concentration\n")
        for a in self.activators:
            txtbc.write("{s} = {v}\n".format(s=a[0], v=a[1]))
        txtbc.write("\n")

        txtbc.write("% Drug Concentration\n")
        txtbc.write("\n")

        txtbc.write("% Time Variables\n")
        for a in self.activators:
            txtbc.write("{s}_on = {time}\n".format(s=a[0], time=a[2]))
        txtbc.write("\n")

        ### Model variables

        txtbc.write("********** MODEL VARIABLES\n")
        txtbc.write("\n")

        # NOTE: deprecated code below, have not deleted for possible later reference

        # for fam in families:

        #     totalStr = "{familyName}_Total = ".format(familyName=fam)
        #     for member in self.specieFamily[fam]: 
        #         totalStr = totalStr + member + " + "

        #     totalStr = totalStr[:-3]
        #     txtbc.write(totalStr)
        #     txtbc.write("\n")

        txtbc.write("\n")
        for a in self.activators:
            txtbc.write("{s}0 = {s}*piecewiseIQM(1,ge(time,{s}_on),0)\n".format(s=a[0]))

        ### Model reactions

        txtbc.write("********** MODEL REACTIONS\n")

        i = 0 
        while i < len(self.reactions):
            re = self.reactions[i]
            id_ = i + 1

            txtbc.write(re.getEqHeaderStr(id_))
            txtbc.write("\n")
            if re.getForwardEqStr(id_) is not None:
                txtbc.write("\tvf = " + re.getForwardEqStr(id_))
                txtbc.write("\n")
            if re.getBackwardEqStr(id_) is not None:
                txtbc.write("\tvr = " + re.getBackwardEqStr(id_))
                txtbc.write("\n")
            
            txtbc.write("\n")
            i += 1

        txtbc.write("********** MODEL FUNCTIONS")
        txtbc.write("\n")
        txtbc.write("\n")
        txtbc.write("\n")

        txtbc.write("********** MODEL EVENTS")
        txtbc.write("\n")
        txtbc.write("\n")
        txtbc.write("\n")
        txtbc.write("\n")

        txtbc.write("********** MODEL MATLAB FUNCTIONS")
        txtbc.write("\n")
        txtbc.write("\n")
        txtbc.write("\n")
        txtbc.write("\n")

        ### END

        txtbc.close()

# if __name__ == "__main__":

#     from MassAction import MassAction
#     from MichaelisMentenGeneral import MichaelisMentenGeneral
#     from Synthesis import Synthesis
#     from Degradation import Degradation

#     peakModel = Model()
#     peakModel.addActivation("EGF", 100, 5000)
#     peakModel.specieConc(['SFKs'],[0])
#     r1 = MichaelisMentenGeneral("EGFR", "uEGFR")
#     r1.addStimulator("PEAK3duCbl")
#     r1.addStimulator("Grb2upPEAK3duCbl")
#     r1.addInhibitor("PEAK3duPYK2")
#     r1.addInhibitor("Grb2upPEAK3duPYK2")
#     r1.noBackward = True

#     r2 = MichaelisMentenGeneral("EGFR", "pEGFR")
#     r2.addStimulator("EGF0", kc=100) # if kc is too low model have no response

#     r3 = MichaelisMentenGeneral("iSFKs", "SFKs")
#     r3.addStimulator("pEGFR", kc=1)

#     r4 = MichaelisMentenGeneral("Shc", "pShc")
#     r4.addStimulator("pEGFR", kc=1)

#     r5 = MassAction("PEAK3", "PEAK3", "PEAK3d")

#     r6 = MassAction("Grb2", "PEAK3d", "Grb2uPEAK3d")

#     r7 = MassAction("PEAK3d", "Cbl", "PEAK3duCbl", Ka=0.01) # Cbl binding with PEAK3 stronger
#     r8 = MassAction("PEAK3d", "PYK2", "PEAK3duPYK2")
#     r9 = MassAction("PEAK3d", "CrkII", "PEAK3duCrkII")

#     r10 = MichaelisMentenGeneral("Grb2uPEAK3d", "Grb2upPEAK3d")
#     r10.addStimulator("SFKs")
#     r10.noBackward = True

#     r11 = MassAction("Grb2uPEAK3d", "pShc", "Grb2uPEAK3dupShc", Ka=0.1, Kd=10)

#     r12 = MassAction("Grb2upPEAK3d", "pShc", "Grb2upPEAK3dupShc", Ka=0.1, Kd=10)

#     r13 = MichaelisMentenGeneral("Grb2upPEAK3dupShc", "Grb2uPEAK3dupShc")
#     r13.addStimulator("PTPN12", kc=100)
#     r13.noBackward = True


#     r14 = MassAction("ASAP1", "Grb2upPEAK3d", "Grb2upPEAK3duASAP1")

#     r15 = MassAction("Grb2upPEAK3d", "Cbl", "Grb2upPEAK3duCbl")
#     r16 = MassAction("Grb2upPEAK3d", "PYK2", "Grb2upPEAK3duPYK2")
#     r17 = MassAction("Grb2upPEAK3d", "CrkII", "Grb2upPEAK3duCrkII", Ka=0.01) # CrkII binding with pPEAK3 stronger

#     r18 = Synthesis("PTPN12", KSyn=0) # PTPN12 stays constant 

#     r19 = Degradation("uEGFR")
#     r20 = Synthesis("EGFR")

#     reactions = [r1, r2, r3, r4, r5, r6, r7, r8,
#                 r9, r10, r11, r12, r13, r14, r15, r16, r17, r18, r19, r20]

#     for r in reactions:
#         peakModel.addReaction(r)

#     peakModel.setModelName("PEAK3-Model-4d")
#     peakModel.generateTxtbc()
#     print(peakModel.species)
#     peakModel.generateTxtbc()


from Reactions import Reactions

class MichaelisMenten(Reactions):

    def __init__(self, forwardSpecie: str, backwardSpecie: str, Vmax_f: float = 10, Km_b: float = 100, Vmax_b: float = 10, Km_f: float = 100):

        super().__init__(forwardSpecie, backwardSpecie)
        self.Vmax_f = Vmax_f
        self.Km_f = Km_f 
        self.Vmax_b = Vmax_b
        self.Km_b = Km_b
        self.type = "MichaelisMenten"

    def computeForward(self, forwardValues: list):

        return (self.Vmax_f * forwardValues[0]) / (self.Km_f + forwardValues[0])

    def computeBackward(self, backwardValues: list):

        return (self.Vmax_b * backwardValues[0]) / (self.Km_b + backwardValues[0])

    def getEqHeaderStr(self, index):
        return "{forward} <=> {backward} :R{i}".format(forward=self.fs[0], backward=self.bs[0], i=index)

    def getForwardEqStr(self, index):
        
        return "Vmax{id}f * {fs} / (Km{id}f + {fs})".format(id=str(index), fs=self.fs[0])

    def getBackwardEqStr(self, index):
        
        return "Vmax{id}b * {bs} / (Km{id}b + {bs})".format(id=str(index), bs=self.bs[0])

    def getParams(self, index):

        return {"Vmax{id}f".format(id=index): self.Vmax_f, "Vmax{id}b".format(id=index): self.Vmax_b, "Km{id}f".format(id=index): self.Km_f, "Km{id}b".format(id=index): self.Km_b}



from Reactions import Reactions

class MassAction(Reactions):

    # Simplified Mass Action with only two forward specie and one backward specie,
    # with molecularities of 1

    def __init__(self, forwardSpecie1: str, forwardSpecie2: str, backwardSpecie: str, Ka: float = 0.001, Kd: float = 0.01):

        super().__init__([forwardSpecie1, forwardSpecie2], backwardSpecie)
        self.Ka = Ka
        self.Kd = Kd
        self.type = "MassAction"

    def computeForward(self, forwardValues: list):

        forwardSpecie1 = forwardValues[0]
        forwardSpecie2 = forwardValues[1]

        return self.Ka * forwardSpecie1 * forwardSpecie2

    def computeBackward(self, backwardValues: list):

        backwardSpecie = backwardValues[0]

        return self.Kd * backwardSpecie

    def getEqHeaderStr(self, index):
        return "{forward1} + {forward2} <=> {backward} :R{i}".format(forward1=self.fs[0], forward2=self.fs[1], backward=self.bs[0], i=index)

    def getForwardEqStr(self, index):
        
        return "Ka{i} * {fs1} * {fs2}".format(i=index, fs1=self.fs[0], fs2=self.fs[1])

    def getBackwardEqStr(self, index):
        
        return "Kd{i} * {bs}".format(i=index, bs=self.bs[0])

    def getParams(self, index):

        return {"Ka{id}".format(id=index): self.Ka, "Kd{id}".format(id=index): self.Kd}
